SQL_HISTORY_RANGE = """
                      SELECT min({key_field}) as min_id, 
                             max({key_field}) as max_id 
                       FROM {name}
                       {where_condition} 
                    """

SQL_HISTORY_RANGE_NO_INT = """
                            SELECT min(__id__) as min_id, max(__id__) as max_id 
                            FROM (
                                SELECT ROW_NUMBER() OVER (ORDER BY {key_field}) as "__id__",
                                                                         *
                                FROM {name}
                                {where_condition} 
                            ) AS subquery
                            """

SQL_HISTORY_NO_INT = """
                            SELECT {fields} 
                            FROM (
                                SELECT ROW_NUMBER() OVER (ORDER BY {key_field}) as "__id__",
                                                                          *
                                FROM {name}
                                {where_condition}
                            ) AS subquery
                            WHERE __id__ BETWEEN '{range_from}' AND '{range_to}'
                            """

SQL_HISTORY = """
                 SELECT {fields} 
                    FROM public.{name} 
                 WHERE {key_field} BETWEEN '{range_from}' AND '{range_to}'
               """

SQL_PG_TABLES = """
                   SELECT table_name
                      FROM information_schema.tables
                   WHERE table_schema = 'public'
                      AND table_type = 'BASE TABLE'
                      AND table_name not in ({tables_to_exclude})
                """

SQL_CH_TABLES = """
                   SELECT name
                       FROM system.tables
                   WHERE database = '{database}'
                        AND name not in ({tables_to_exclude})
                """

SQL_CH_SCHEMA = """SELECT name, type, comment 
                   FROM system.columns 
                   WHERE database='{database}' AND table='{table}'
                         AND name NOT IN ({exclude_columns})   
                """

SQL_SORT_PART = """SELECT DISTINCT sorting_key, partition_key
                    FROM system.tables
                    WHERE name = '{name}'
                          AND database = '{database}'
                """

SQL_CREATE_AS = """CREATE TABLE {staging_database}.{temp_table_prefix}{table} ON CLUSTER '{{cluster}}' AS {database}.{table} 
                   ENGINE=ReplicatedMergeTree() 
                   PARTITION BY {temp_table_partition_by} ORDER BY ({temp_table_order_by})
                """

SQL_DROP_TABLE = """DROP TABLE IF EXISTS {database}.{prefix}{table} ON CLUSTER '{{cluster}}'"""

SQL_PARTS = """SELECT partition, count() cnt 
               FROM system.parts 
               WHERE database='{database}' AND table='{table}' and active
               GROUP BY partition 
                 HAVING cnt > 1
            """

SQL_EXCHANGE_TABLE = """EXCHANGE TABLES {database_staging}.{prefix}{table} AND {database}.{table} ON CLUSTER '{{cluster}}' """


import logging
import sys
from pg_s3_ch.sql_queries import SQL_CH_TABLES, SQL_PG_TABLES
from pg_s3_ch.pg_s3_ch import PGS3CH


def setup_logging():
    global logger
    logger = logging.getLogger("pg_s3_ch")
    logging.basicConfig(
        stream=sys.stdout,
        level="INFO",
        format="%(asctime)s %(processName)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


class Syncer(PGS3CH):

    def __init__(self, ch_config, pg_config):
        super().__init__(
            config={},
            s3_config={},
            ch_config=ch_config,
            pg_config=pg_config
        )

    def sync_tables(self, tables_to_exclude=None):
        """
        :param tables_to_exclude: list of tables shouldn't  being considered
        :return: tuple : (list of tables need to remove, list of tables need to create or recreate)
        """
        logger.info("Get tables from pg")
        cnx = self.connect_to_pg()

        cursor = cnx.cursor()
        tables_to_exclude = (
            ",".join(["'" + item + "'" for item in tables_to_exclude.split(",")])
            if tables_to_exclude
            else "''"
        )

        sql = SQL_PG_TABLES.format(tables_to_exclude=tables_to_exclude)
        cursor.execute(sql)
        tables_pg = [r[0] for r in cursor]
        cursor.close()
        cnx.close()

        logger.info("Get tables from ch")

        sql = SQL_CH_TABLES.format(
            database=self.ch_config["database"], tables_to_exclude=tables_to_exclude
        )
        result = self.ch_client.execute(sql)
        tables_ch = [r[0] for r in result]
        return tables_ch, tables_pg

    def save_sync_tables(self, client_description, save_table_name, tables_to_exclude):
        """

        :param client_description:
        :param save_table_name:
        :param tables_to_exclude:
        :return: None
        """
        tables_ch, tables_pg = self.sync_tables(tables_to_exclude=tables_to_exclude)
        tables_to_remove = [c for c in tables_ch if c not in tables_pg]

        new_tables = []
        for row in tables_pg:
            items = {
                "client_description": client_description,
                "table_name_ch": row,
                "table_name_pg": row,
                "database_pg": self.pg_config["database"],
                "database_ch": self.ch_config["database"],
            }

            new_tables.append(items)

        table_name = f"{save_table_name}"
        if len(tables_pg) > 0:
            columns = list(items.keys())
            columns = ",".join(columns)

        for items in new_tables:

            sql_stat = " SELECT 1 FROM {table_name}" \
                       " WHERE database_pg = '{database_pg}' " \
                       " AND table_name_pg = '{table_name_pg}' " \
                       " AND database_ch = '{database_ch}'".format(
                table_name=table_name,
                database_pg=self.pg_config["database"],
                table_name_pg=items["table_name_pg"],
                database_ch=self.ch_config["database"],
            )

            check_statement = self.ch_client.execute(sql_stat)
            if len(check_statement) == 0:
                logger.info("Add new table %s", table_name)
                self.ch_client.execute(
                    f"INSERT INTO {table_name} ({columns}) VALUES",
                    [items],
                    types_check=True,
                )

        for items in tables_to_remove:
            drop_stat = "DROP TABLE IF EXISTS  {database_ch}.{table_name_pg}".format(
                database_ch=self.ch_config["database"], table_name_pg=items
            )
            logger.info(drop_stat)



import re
from sqlalchemy.orm import Query, Session


class Compiler(object):
    """Supporting code for tests"""

    strip_spaces = re.compile(r"[\n\t]")

    def __init__(self, db_worker):
        self.db_worker = db_worker
        self.session = Session(self.db_worker.engine)

    def _compile(
            self, clause, bind=None, literal_binds=False, render_postcompile=False
    ):
        if bind is None:
            bind = self.session.bind

        if isinstance(clause, Query):
            context = clause._compile_context()
            clause = context.query

        kw = {}
        compile_kwargs = {}
        if literal_binds:
            compile_kwargs["literal_binds"] = True
        if render_postcompile:
            compile_kwargs["render_postcompile"] = True

        if compile_kwargs:
            kw["compile_kwargs"] = compile_kwargs

        return clause.compile(dialect=bind.dialect, **kw)

    def compile(self, clause, **kwargs):
        return self.strip_spaces.sub("", str(self._compile(clause, **kwargs)))



import os
import sys
import math
from datetime import datetime, timedelta, time
from typing import Optional
from enum import Enum

import logging
import gzip
import psycopg2
import clickhouse_driver
import psycopg2.extras
from s3.client import Client as s3_client
from clickhouse_driver import Client
from .sql_queries import (
    SQL_HISTORY,
    SQL_HISTORY_RANGE,
    SQL_HISTORY_RANGE_NO_INT,
    SQL_HISTORY_NO_INT,
    SQL_CH_SCHEMA,
    SQL_PARTS,
)
from .temp_table_creator_ch import TempTableCreatorCH

PG_HISTORY_STEP_DEFAULT = 500000
FILE_PATH = "./csv/{entity_name}_{range_from}.csv.gz"
S3_FILE_FORMAT = "{endpoint_url}/{bucket}/{upload_path}/{entity_name}_*.csv.gz"
EXPORT_PATH = "./csv"


def setup_logging():
    global logger
    logger = logging.getLogger("pg_s3_ch")
    logging.basicConfig(
        stream=sys.stdout,
        level="INFO",
        format="%(asctime)s %(processName)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def create_files_dir():
    try:
        logging.info("Working directory is {getcwd}".format(getcwd=os.getcwd()))
        os.makedirs(EXPORT_PATH)
    except OSError:
        logging.info(
            "Can't create directory {export_path}".format(export_path=EXPORT_PATH)
        )
    else:
        logging.info(
            "{export_path} was successfully created".format(export_path=EXPORT_PATH)
        )


def drop_files_dir():
    if os.path.isdir(EXPORT_PATH):
        for file in os.listdir(EXPORT_PATH):
            logging.info("Remove file %s", file)
            os.remove(EXPORT_PATH + "/" + file)
        os.removedirs(EXPORT_PATH)


def sql_to_file(cnx, file_path, sql):
    """
    Save result of sql to file
    :param cnx:
    :param file_path:
    :param sql:
    :return:
    """
    with gzip.open(file_path, "wt") as fa:
        cursor = cnx.cursor()
        cursor.copy_to(fa, "({sql})".format(sql=sql), null="")
        cursor.close()

    # with gzip.open(file_path, 'wt') as fa:
    #     cursor = cnx.cursor()
    #     sql_copy = "COPY ({sql}) TO STDOUT WITH NULL AS '' csv DELIMITER '\t' "
    #     cursor.copy_expert(sql_copy.format(sql=sql), file=fa)
    #
    #     cursor.close()


def file_to_s3(s3_client_, file_path, s3_path):
    """
    Save result of sql to file
    :param s3_path:
    :param s3_client_:
    :param file_path:
    :return:
    """
    logging.info(os.path.basename(file_path))
    s3_client_.upload_file(file_path, s3_path)


def tmp_table_wrapper(f):
    def wrapper(*args):
        with TempTableCreatorCH(
                ch_client=args[0].ch_client,
                table_name=args[0].ch_name,
                table_name_tmp=args[0].temp_table_prefix + args[0].ch_name,
                database=args[0].ch_config["database"],
                staging_database=args[0].ch_config["staging_database"],
                sorting_key=args[0].sorting_key,
                partition_key=args[0].partition_key,
                ttl=args[0].ttl
        ) as tmp_creator:
            f(*args)
            if not args[0].by_date:
                tmp_creator.exchange()

    return wrapper


def check_ch_config(ch_config: dict):
    for param in ["host", "port", "user", "password", "database", "staging_database"]:
        if param not in ch_config:
            raise BaseException(
                "Error. You must specify {} param in ch_config".format(param)
            )


def check_pg_config(pg_config: dict):
    for param in ["host", "port", "user", "password", "database"]:
        if param not in pg_config:
            raise BaseException(
                "Error. You must specify {} param in ch_config".format(param)
            )


def check_s3_config(s3_config: dict):
    for param in [
        "S3_ENDPOINT_URL",
        "S3_TOPMIND_CLIENT_DATA_BUCKET",
        "S3_FOLDER",
        "S3_ACCESS_KEY",
        "S3_ACCESS_SECRET",
        "UPLOAD_PATH",
    ]:
        if param not in s3_config:
            raise BaseException(
                "Error. You must specify {} param in s3_config".format(param)
            )


def check_config(config: dict):
    for param in ["name"]:
        if param not in config:
            raise BaseException(
                "Error. You must specify {} param in config".format(param)
            )


def set_bool_value(config, key, default_value):
    if config.get(key, default_value) is None:
        return default_value
    else:
        return bool(int(config.get(key, default_value)))


class TypeUpload(Enum):
    BATCH = 0
    BY_DATE = 1
    INCREMENTAL = 2


class PGS3CH:
    """
      Represents class realizing the logic of transferring data from PostgreSQL to click house
      :param config:
      :param s3_config:
      :param ch_config:
      :param pg_config:
      """

    def __init__(self, config, s3_config, ch_config, pg_config):

        setup_logging()

        check_config(config)
        check_s3_config(s3_config)
        check_ch_config(ch_config)

        self.config = config
        self.s3_config = s3_config
        self.pg_config = pg_config
        self.ch_config = ch_config

        self.execution_date = self.config.get("execution_date") or os.getenv("execution_date")

        self.entity_name: str = ""  # table name in pg
        self.ch_name: str = ""  # table name in ch
        self.partition_key: str = "tuple()"
        self.sorting_key: str = "id"
        self.key_field: str = "id"
        self.pg_history_step: int = PG_HISTORY_STEP_DEFAULT
        self.temp_table_prefix: str = "_temp_"
        self.ttl: Optional[str] = ""

        self.where_condition: Optional[str] = None
        self.incremental: Optional[bool] = None
        self.incremental_field: Optional[str] = None
        self.range_parts_per_day: Optional[
            int
        ] = None  # Optional if incremental or by date
        self.by_date: Optional[bool] = None  # Optional if download by date
        self.date_field: Optional[str] = None  # Optional
        self.need_clear_s3: Optional[bool] = None  #
        self.need_optimize: Optional[bool] = None

        self.ch_exclude_columns_str: str = ""
        self.ch_exclude_columns: Optional[
            list[str]
        ] = None  # Columns we should ignore in clickhouse table

        self.type_upload: TypeUpload = TypeUpload.BATCH  # Type upload,  default Batch
        self.transform_params()  # get transformed params from the config

        self.ch_client = None
        self.connect_to_ch()

        self.schema_ch_insert_fields = []
        self.schema_ch_insert_names = []
        self.schema_pg_select_fields = {}
        self.ch_fields = {}
        self.key_field_is_int = True

    def transform_params(self):

        self.entity_name = self.config.get("name")
        self.ch_name = self.config.get("ch_name", self.entity_name) or self.entity_name
        self.key_field = self.config.get("key_field", "id") or "id"
        self.pg_history_step = int(
            self.config.get("pg_history_step", PG_HISTORY_STEP_DEFAULT)
            or PG_HISTORY_STEP_DEFAULT
        )
        self.need_optimize = set_bool_value(self.config, "need_optimize", True)
        self.incremental = set_bool_value(self.config, "incremental", False)
        self.need_clear_s3 = set_bool_value(self.config, "need_clear_s3", False)
        self.by_date = set_bool_value(self.config, "by_date", False)

        self.partition_key = self.config.get("partition_key", "tuple()") or "tuple()"
        self.sorting_key = self.config.get("sorting_key", "id") or "id"
        self.ttl = self.config.get("ttl", "") or ""
        self.temp_table_prefix = self.ch_config.get("temp_table_prefix", "_temp_") or "_temp_"

        if self.by_date:
            self.type_upload = TypeUpload.BY_DATE
            self.temp_table_prefix += self.execution_date[0:10].replace("-", "")
            self.date_field = self.config.get("date_field", "created_at") or "created_at"
            self.incremental_field = self.date_field
            self.range_parts_per_day = int(self.config.get("range_parts_per_day", 24) or 24)

        elif self.incremental:
            self.type_upload = TypeUpload.INCREMENTAL
            self.range_parts_per_day = int(self.config.get("range_parts_per_day", 24) or 24)
            self.incremental_field = str(
                self.config.get("incremental_field", "updated_at") or "updated_at"
            )

        self.ch_exclude_columns_str = self.ch_config.get(
            "ch_exclude_columns", "ts_captured"
        ) or "ts_captured"

        self.ch_exclude_columns = ",".join(
            ["'" + item + "'" for item in self.ch_exclude_columns_str.split(",")]
        )

        self.where_condition = str(self.config.get("where_condition", "") or "")

    def get_table_schema_ch(self):
        sql = SQL_CH_SCHEMA.format(
            database=self.ch_config["database"],
            table=self.ch_name,
            exclude_columns=self.ch_exclude_columns,
        )

        logger.info(sql)
        result = self.ch_client.execute(sql)
        try:
            type_key_field = [r[1] for r in result if r[0] == self.key_field][0]
        except IndexError as E:
            logging.error('Can not find key field. Even "id" is not in schema')
            raise IndexError(E)

        if "int" not in type_key_field.lower():
            self.key_field_is_int = False

        for row in result:
            if row[0].replace("_dt", "") in self.ch_exclude_columns_str.split(","):
                continue
            if row[2].find("Exclude") == -1:
                self.schema_pg_select_fields[row[0]] = row[1]

            if row[2].find("#OnInsert:") > -1:
                start = row[2].find("#OnInsert:") + 10
                end = row[2].find(":EndOnInsert#")
                self.schema_ch_insert_fields.append(row[2][start:end])
            else:
                self.schema_ch_insert_fields.append(row[0])
            self.schema_ch_insert_names.append(row[0])

        if len(self.schema_pg_select_fields) < 1:
            logger.info("CH schema error: %s ", self.schema_pg_select_fields)
            sys.exit(1)

        if len(self.schema_ch_insert_fields) < 1:
            logger.info(
                "CH schema error: %s. Have you already created the table?", self.ch_name
            )

            sys.exit(1)

        logger.info(
            "CH schema for %s.%s - OK", self.ch_config["database"], self.ch_name
        )

    def connect_to_ch(self):

        settings = {"insert_quorum": os.getenv('REPLICAS_NUMBERS') or 3}

        self.ch_client = Client(
            host=self.ch_config["host"],
            user=self.ch_config["user"],
            password=self.ch_config["password"],
            database=self.ch_config["database"],
            settings=settings,
        )

    def clear_s3_folder(self):
        """Clear all data from s3 folder"""
        client = s3_client(
            aws_access_key_id=self.s3_config["S3_ACCESS_KEY"],
            aws_secret_access_key=self.s3_config["S3_ACCESS_SECRET"],
            endpoint_url=self.s3_config["S3_ENDPOINT_URL"],
            bucket=self.s3_config["S3_TOPMIND_CLIENT_DATA_BUCKET"],
        )

        s3_file_path = S3_FILE_FORMAT.format(
            endpoint_url=self.s3_config["S3_ENDPOINT_URL"],
            bucket=self.s3_config["S3_TOPMIND_CLIENT_DATA_BUCKET"],
            upload_path=self.s3_config["UPLOAD_PATH"],
            entity_name=self.entity_name,
        )

        if client.path_exists(s3_file_path):
            for file in self.s3_client.get_file_list(self.s3_client.root_dir):
                self.logger.info("Delete %s from s3", file)
                self.s3_client.delete_file(path=file)
            self.s3_client.delete_dir(self.s3_client.root_dir)  # remove directory

    def connect_to_pg(self):
        cnx = psycopg2.connect(
            user=self.pg_config["user"],
            password=self.pg_config["password"],
            host=self.pg_config["host"],
            port=self.pg_config["port"],
            database=self.pg_config["database"],
            sslmode="disable",
            connect_timeout=10,
        )

        cnx.set_client_encoding("UTF8")
        return cnx

    def extract_to_s3(self):
        logger.info("Starting extract to s3")
        cnx = self.connect_to_pg()
        cursor = cnx.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
        sql_history_local = SQL_HISTORY

        if self.key_field_is_int:
            sql = SQL_HISTORY_RANGE.format(
                key_field=self.key_field,
                name=self.entity_name,
                where_condition="WHERE " + self.where_condition
                if self.where_condition
                else "",
            )
        else:
            logger.info("Change logic of extracting data")
            sql = SQL_HISTORY_RANGE_NO_INT.format(
                key_field=self.key_field,
                name=self.entity_name,
                where_condition="WHERE " + self.where_condition
                if self.where_condition
                else "",
            )

            sql_history_local = SQL_HISTORY_NO_INT

        cursor.execute(sql)

        for row in cursor:
            min_id = row["min_id"]
            max_id = row["max_id"]
        cursor.close()

        logger.info("Min ID: %d", min_id)
        logger.info("Max ID: %d", max_id)

        if min_id is None or max_id is None:
            logger.info("Max_id or min_id is None. Process is stopped.")
            sys.exit(0)

        step = self.pg_history_step

        end = math.ceil((max_id - min_id) / step)
        logger.info("Range from 0 to %s, multiply by %s", end, step)

        client = s3_client(
            aws_access_key_id=self.s3_config["S3_ACCESS_KEY"],
            aws_secret_access_key=self.s3_config["S3_ACCESS_SECRET"],
            endpoint_url=self.s3_config["S3_ENDPOINT_URL"],
            bucket=self.s3_config["S3_TOPMIND_CLIENT_DATA_BUCKET"],
        )
        create_files_dir()

        for _id in range(0, end):
            range_from = min_id + _id * step
            range_to = min_id + _id * step + step - 1

            sql = sql_history_local.format(
                key_field=self.key_field,
                name=self.entity_name,
                fields='"' + '","'.join(self.schema_pg_select_fields.keys()) + '"',
                range_from=range_from,
                range_to=range_to,
                where_condition="WHERE " + self.where_condition
                if self.where_condition
                else "",
            )

            logger.info(sql)
            file_path = FILE_PATH.format(
                entity_name=self.entity_name, range_from=range_from
            )
            logger.info("Saving local: %s", file_path)
            sql_to_file(cnx, file_path, sql)

            path = "{upload_path}/{filename}".format(
                upload_path=self.s3_config["UPLOAD_PATH"],
                filename=os.path.basename(file_path),
            )
            logger.info(
                "Copying to S3: %s", self.s3_config["S3_ENDPOINT_URL"] + "/" + path
            )
            file_to_s3(client, file_path, path)

        cnx.close()
        drop_files_dir()
        logger.info("Full extract finished")

    def optimize(self):
        """
        Optimize table
        :return: None
        """
        sql = SQL_PARTS.format(database=self.ch_config["database"], table=self.ch_name)
        result = self.ch_client.execute(sql)

        for row in result:
            logger.info("Partition %s has %s parts", row[0], row[1])

            sql_optimize = """OPTIMIZE TABLE {database}.{table} ON CLUSTER '{{cluster}}' PARTITION {partition} """.format(
                database=self.ch_config["database"],
                table=self.ch_name,
                partition=row[0],
            )
            logger.info(sql_optimize)
            try:
                self.ch_client.execute(sql_optimize)
            except clickhouse_driver.errors.ServerException as E:
                logging.info(str(E))

            logger.info("OK")

    def s3_to_temp(self):
        """

        :return:
        """

        schema = []
        schema_insert = []

        for k in self.schema_pg_select_fields:
            schema.append(k + " " + self.schema_pg_select_fields[k])

        for k in self.schema_ch_insert_fields:
            schema_insert.append(k)

        logger.info(
            "Copying from %s/%s to %s.%s%s",
            self.s3_config["S3_ENDPOINT_URL"],
            self.s3_config["S3_TOPMIND_CLIENT_DATA_BUCKET"],
            self.ch_config["database"],
            self.temp_table_prefix,
            self.ch_name,
        )
        s3_file_path = S3_FILE_FORMAT.format(
            endpoint_url=self.s3_config["S3_ENDPOINT_URL"],
            bucket=self.s3_config["S3_TOPMIND_CLIENT_DATA_BUCKET"],
            upload_path=self.s3_config["UPLOAD_PATH"],
            entity_name=self.entity_name,
        )

        sql = """
            INSERT INTO {staging_database}.{temp_table_prefix}{table}  ({schema_insert_name})
            SELECT 
                {schema_insert}, NOW()
            FROM s3('{s3_file_path}', 
                    '{S3_ACCESS_KEY}',
                    '{S3_ACCESS_SECRET}',
                    'TSV', 
                    '{schema}',
                    'gzip'
                    );
            """.format(
            s3_file_path=s3_file_path,
            schema_insert_name='"'
                               + '","'.join(self.schema_ch_insert_names + ["ts_captured"])
                               + '"',
            staging_database=self.ch_config["staging_database"],
            table=self.ch_name,
            temp_table_prefix=self.temp_table_prefix,
            S3_ACCESS_KEY=self.s3_config["S3_ACCESS_KEY"],
            S3_ACCESS_SECRET=self.s3_config["S3_ACCESS_SECRET"],
            entity_name=self.entity_name,
            schema=", ".join(schema),
            schema_insert=", ".join(schema_insert),
        )
        logger.info(sql)

        self.ch_client.execute(sql)

        logger.info("Loading from S3 successful")

    def upload_temp_to_prod(self):
        """

        :return:
        """
        "INSERT INTO "
        sql = f"INSERT INTO {self.ch_config['database']}.{self.ch_name} SELECT * FROM {self.ch_config['staging_database']}.{self.temp_table_prefix}{self.ch_name}"
        logger.info(sql)
        self.ch_client.execute(sql)

    def check_date_exists(self):
        sql = """
                SELECT COUNT(*)
                FROM {database}.{table}   
                WHERE toDate({date_field}) = '{date}'
              """.format(
            database=self.ch_config["database"],
            table=self.ch_name,
            date_field=self.date_field + "_dt",
            date=self.execution_date[0:10],
        )
        logging.info(sql)
        return self.ch_client.execute(sql)[0][0]

    def extract_updated_to_s3(self):
        """
        Extract incremental data
        :return: None
        """
        cnx = self.connect_to_pg()
        cursor = cnx.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        range_from = datetime.combine(
            datetime.strptime(self.execution_date[0:10], "%Y-%m-%d").date(), time()
        )

        parts = self.range_parts_per_day

        client = s3_client(
            aws_access_key_id=self.s3_config["S3_ACCESS_KEY"],
            aws_secret_access_key=self.s3_config["S3_ACCESS_SECRET"],
            endpoint_url=self.s3_config["S3_ENDPOINT_URL"],
            bucket=self.s3_config["S3_TOPMIND_CLIENT_DATA_BUCKET"],
        )

        create_files_dir()

        for _ in range(0, parts):
            range_to = range_from + timedelta(seconds=3600 * 24 / parts - 0.000001)
            if range_from > datetime.now() + timedelta(hours=3):
                break

            sql = SQL_HISTORY.format(
                key_field=self.incremental_field,
                name=self.entity_name,
                fields='"' + '","'.join(self.schema_pg_select_fields.keys()) + '"',
                range_from=range_from,
                range_to=range_to,
            )
            logger.info(sql)
            file_path = FILE_PATH.format(
                entity_name=self.entity_name, range_from=range_from
            )
            logger.info("Saving local: %s", file_path)
            sql_to_file(cnx, file_path, sql)

            path = "{upload_path}/{filename}".format(
                upload_path=self.s3_config["UPLOAD_PATH"],
                filename=os.path.basename(file_path),
            )
            logger.info(
                "Copying to S3: %s", self.s3_config["S3_ENDPOINT_URL"] + "/" + path
            )
            file_to_s3(client, file_path, path)

            range_from = range_from + timedelta(hours=24 / parts)

        drop_files_dir()
        cursor.close()
        cnx.close()

        logging.info("extract_updated finished")

    def copy_prod_to_temp(self):
        """
        Insert all data except incrememtal to temp table
        :return: None
        """
        sql = """
                INSERT INTO {staging_database}.{temp_table_prefix}{table} 
                SELECT * FROM {database}.{table}   
                WHERE {primary_key_field} 
                   NOT IN (SELECT {primary_key_field} 
                           FROM {staging_database}.{temp_table_prefix}{table})
              """.format(
            staging_database=self.ch_config["staging_database"],
            temp_table_prefix=self.temp_table_prefix,
            database=self.ch_config["database"],
            table=self.ch_name,
            primary_key_field=self.key_field,
        )

        logging.info(sql)
        self.ch_client.execute(sql)

    @tmp_table_wrapper
    def transfer_data_incremental(self):
        """
        Transfer data incremental
        :return: None
        """
        if not self.incremental:
            raise ValueError("Incremental parameter have to set True or 1")

        self.get_table_schema_ch()
        self.extract_updated_to_s3()
        self.s3_to_temp()
        self.copy_prod_to_temp()

    @tmp_table_wrapper
    def transfer_data_full(self):
        """
        Transfer full data
        :return: None
        """
        self.get_table_schema_ch()
        self.extract_to_s3()
        self.s3_to_temp()

    @tmp_table_wrapper
    def transfer_data_by_date(self):
        """
        Transfer data by date
        :return: None
        """
        if self.check_date_exists() > 0:
            raise BaseException(
                f"Date {self.execution_date[0:10]} is already exists in the table"
            )

        self.get_table_schema_ch()
        self.extract_updated_to_s3()
        self.s3_to_temp()
        self.upload_temp_to_prod()

    def transfer(self):
        if self.incremental:
            self.transfer_data_incremental()
        elif self.by_date:
            self.transfer_data_by_date()
        else:
            self.transfer_data_full()

        if self.need_optimize:
            self.optimize()



from sqlalchemy import Column, func
from clickhouse_sqlalchemy import types

PARSE_DATE_COLUMN = (
    "#Exclude #OnInsert:parseDateTimeBestEffortOrZero({column}):EndOnInsert#"
)

CUSTOM_TYPES = {
    "timestamp": "String",
    "timestamp without time zone": "String",
    "timestamp with time zone": "String",
    "time without time zone": "String",
    "date": "String",
    "datetime": "String",
}


def get_timestamp_columns(pg_columns: list) -> list:
    """
    Get timestamp columns,
    :param pg_columns:
    :return:
    """
    return list(c["column_name"] for c in pg_columns if c["data_type"] == "timestamp" or c["data_type"] == "date")


def insert_new_columns(pg_columns: list, columns: list[Column]):
    """
    Insert custom columns to new table
    :param pg_columns:
    :param columns:
    :return:
    """
    timestamp_columns = get_timestamp_columns(pg_columns=pg_columns)
    for column in columns.copy():
        if column.name in timestamp_columns:
            new_column = Column(
                name=column.name + "_dt",
                type_=types.DateTime,
                comment=PARSE_DATE_COLUMN.format(column=column.name),
            )
            columns.insert(columns.index(column) + 1, new_column)
    # ts_captured as last column
    dt = Column("ts_captured", types.DateTime, server_default=func.now())
    columns.append(dt)


def exclude_columns(columns: list) -> list:
    """
    Exclude columns with #Exclude in comments
    :param columns:
    :return:
    """
    for c in columns.copy():
        if "#Exclude" in (c["comment"] or ""):
            columns.remove(c)



import logging

from sqlalchemy.sql.ddl import CreateTable

from converter.pg_ch_converter.schemas_checker import SchemaChecker, ChangeEvent
from .table_transformer import (
    insert_new_columns,
    exclude_columns,
    CUSTOM_TYPES,
)
from .compile_dialect import Compiler


def format_config(config: dict) -> dict:
    columns = ["host", "user", "password", "port", "database"]
    return {item: value for item, value in config.items() if item in columns}


def create_or_update(
        table_name_pg: str,
        table_name_ch: str,
        config: dict,
        pg_config: dict,
        ch_config: dict,
):
    _sql_credentials = {
        "pg": format_config(pg_config),
        "ch": format_config(ch_config),
    }
    sorting_key = config.get("sorting_key", "id") or "id"
    partition_key = config.get("partition_key", "tuple()") or "tuple()"
    ttl = config.get("ttl", "") or ""
    exclude_columns_ch = config.get(
        "ch_exclude_columns", "ts_captured"
    ) or "ts_captured"

    checker = SchemaChecker(
        sql_credentials=_sql_credentials,
        table_name_pg=table_name_pg,
        table_name_ch=table_name_ch,
        sorting_key=sorting_key + " " + ttl,
        partition_key=partition_key,
        custom_types=CUSTOM_TYPES,
        exclude_columns_ch=exclude_columns_ch
    )
    if not checker.db_worker_to.is_table_exists(table_name_ch):
        new_table = checker.create_table(insert_func=insert_new_columns)
        new_table.create()
        test_class = Compiler(checker.db_worker_to)
        logging.info(test_class.compile(CreateTable(new_table)))

    else:
        events = checker.get_difference_events(exclude_function=exclude_columns)
        if len(events):
            for event in events:
                if event.change_event != ChangeEvent.TYPE_UPDATE:
                    checker.db_worker_to.execute(checker.alter_statement(event))
                logging.info(checker.alter_statement(event))



import logging
from typing import Optional
import clickhouse_driver
from clickhouse_driver import Client

logger = logging.getLogger(__name__)


class TempTableCreatorCH(object):
    """create temporary table in clickhouse based on table_name"""

    def __init__(self,
                 ch_client: Client,
                 table_name: str,
                 table_name_tmp: str,
                 database: str,
                 staging_database: str,
                 sorting_key: Optional[str] = 'id',
                 partition_key: Optional[str] = 'tuple()',
                 ttl: Optional[str] = ''):
        """
        :param table_name:
        :param table_name_tmp:
        :param database:
        :param staging_database:
        :param sorting_key:
        :param partition_key:
        :param ttl:
        """
        self.ch_client = ch_client
        self.table_name = table_name
        self.table_name_tmp = table_name_tmp
        self.database = database
        self.staging_database = staging_database
        self.sorting_key = sorting_key
        self.partition_key = partition_key
        self.ttl = ttl

    def create(self):
        statement = f"""CREATE TABLE {self.staging_database}.{self.table_name_tmp} ON CLUSTER '{{cluster}}' AS {self.database}.{self.table_name}
           ENGINE=ReplicatedMergeTree()
           PARTITION BY {self.partition_key} ORDER BY ({self.sorting_key})  {self.ttl}
        """
        logging.info(statement)
        self.ch_client.execute(statement)

    def drop(self):
        statement = f"""DROP TABLE IF EXISTS {self.staging_database}.{self.table_name_tmp} ON CLUSTER '{{cluster}}'"""
        logging.info(statement)
        self.ch_client.execute(statement)

    def exchange(self):
        """
        exchange
        :return:
        """
        statement = f"""EXCHANGE TABLES {self.staging_database}.{self.table_name_tmp} AND {self.database}.{self.table_name} ON CLUSTER '{{cluster}}' """
        logging.info(statement)
        self.ch_client.execute(statement)

    def __enter__(self):
        self.drop()
        self.create()
        return self

    def __exit__(self, type, value, traceback):
        try:
            self.drop()
        except clickhouse_driver.errors.ServerException as E:
            logger.info(str(E))




from setuptools import setup, find_packages


def load_requirements(file_name):
    requirements = []
    with open(file_name, "r") as file:
        for line in file:
            # Remove comments and strip whitespace
            line = line.split("#")[0].strip()
            if line:
                requirements.append(line)
    return requirements


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pg-s3-ch',
    packages=find_packages(),
    version='1.0.1',
    license='MIT',
    description='PG S3 CH Helper',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Dmitry Utiralov',
    author_email='d.utiralov@netology.tech',
    url='https://github.com/severgroup-tt/topmind-commons',
    install_requires=load_requirements("requirements.txt"),
    extras_require={
        "dev": load_requirements("dev-requirements.txt"),
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.9'
)
